import importlib.util as _ilu
import importlib as _il
import os
from os.path import exists
import json
from typing import Any


def module_path(path: str) -> str:
    """Takes in a path that can be either a regular path, a module, or a module followed by a path.
    They can take the form:

        <module>:<path>
        <module>
        <path>

    For example, if the module `sl.module` exists at `/home/user/proj/src/sl/module`, the following
    are example paths and what they would map to:

        # Regular absolute path
        > module_path('/home/user/proj/src/sl/module/subfolder/file.txt')
        '/home/user/proj/src/sl/module/subfolder/file.txt'

        # Module followed by a path
        > module_path('sl.module:subfolder/file.txt')
        '/home/user/proj/src/sl/module/subfolder/file.txt'

        # The path portion can start anywhere, even if it includes a "module" folder in it
        > module_path('sl:module/subfolder/file.txt')
        '/home/user/proj/src/sl/module/subfolder/file.txt'

        # Path directly to the module
        > module_path('sl.module')
        '/home/user/proj/src/sl/module'

        # Relative paths are unchanged
        > module_path('../some_file.txt')
        '../some_file.txt'

    It will not change paths that do not include the module component. If a path does not include
    the colon and has no slashes it will attempt to load the path as a module, and otherwise it
    will return the path unchanged.
    """
    if ":" in path:
        # If a colon does exist then we know for a fact that the lefthand side is meant to be a python module
        module, append = path.split(":")
        spec = _ilu.find_spec(module)
        if not spec or not spec.origin:
            raise ValueError(f"Module `{module}` does not appear to exist")
        dir_path = os.path.dirname(spec.origin)
        # Return the directory path plus the apended value
        return os.path.join(dir_path, os.path.normpath(append))
    # If there are no slashes and no colon, attempt to read the path as a module
    elif not ("/" in path or "\\" in path):
        spec = _ilu.find_spec(path)
        if not spec or not spec.origin:
            raise ValueError(f"Module `{path}` does not appear to exist")
        return os.path.dirname(spec.origin)

    # Otherwise this does not conform to anything we recognize; return the path unchanged
    return path


def import_from_str(module_name):
    """To import a module, just write it as you would a regular module:
        `my.module.path`
    To import a specific variable/function from the module, include the
    variable/function name after a colon:
        `my.module.path:var_name`
    """
    if ":" in module_name:
        module_str, var_str = module_name.split(":")
        module = _il.import_module(module_str)
        return module.__dict__[var_str]
    else:
        return _il.import_module(module_name)


def json_to_str(val: Any, *, sort_keys: bool = True, indent: int = 2) -> str:
    """Dumps to json string using default values. Primarily used for
    writing to files. `val` should already be json-serializable
    """
    return json.dumps(val, sort_keys=sort_keys, indent=indent)


def str_to_json(val: str) -> dict[Any, Any]:
    return json.loads(val)



import typing as _t
import pydantic as _pyd
import fastapi as _fapi


class ApiError(_pyd.BaseModel):
    result: _t.Literal["Err"] = "Err"
    message: str

    def to_httperr(self):
        return _fapi.HTTPException(status_code=404, detail=self.message)


class Credentials(_pyd.BaseModel):
    username: str = _pyd.Field(
        title="Username",
        description="Database username",
    )
    password: str = _pyd.Field(
        title="Password",
        description="Database password",
    )


class CreatedDb(_pyd.BaseModel):
    url: str = _pyd.Field(
        title="URL",
        description="URL string to use to connect to the newly-created database",
    )
    drop_id: str = _pyd.Field(
        title="Drop ID", description="Use this value to drop the database later."
    )


_schemadef_type_desc = """\
Specifies the type of `value`:
- `sqlalchemy` - A path to a sqlalchemy metadata object; eg: `sl.module.db:metadata`.
- `file` - The path to a SQL file that should be run to build the models. It is recommended that
this be an absolute path, and the file must be resident on the same filesystem on which the server
is running. eg: `/path/to/file.sql`.
- `raw_sql` - Raw SQL passed in through the value which will create the schema.
"""


class SchemaDef(_pyd.BaseModel):
    """Define how to load the schema"""

    type: _t.Literal["sqlalchemy", "file", "raw_sql"] = _pyd.Field(
        title="Type", description=_schemadef_type_desc
    )
    value: str = _pyd.Field(
        title="Value",
        description=(
            "Value that corresponds to the type. See the type field for more information."
        ),
    )


_seeddata_type_desc = """\
Specifies the type of `value`:
- `json` - Raw JSON-formatted data, structured as:
```
{
    "metadata": "abs.path.to.module:metadata",
    "data": {
        "table_1_name": [
            {
                "int_col": 1,
                "str_col": "value",
                ...
            },
            ...entries
        ],
        "table_2_name": [ ...entries ],
        ...tables
    }
}
```
- `sql` - Raw SQL containing `INSERT`/`UPDATE`/etc. statements.
- `file` - Path to file local to the dbserver containing `json`/`sql` data structured as described
above
- `module` - Module path to a file local to the dbserver containing `json`/`sql` data structured
as described above. A module path consists of a python module, followed by a colon, then the rest
of the path to the file. eg. `sl.dbserver.__test__:seeds/test.json`. The module path portion can
only include directories recognizable as Python modules (ie. containing an `__init__.py` file).
Any subdirectories after that point which are not python modules must be included after the colon.
"""


class SeedData(_pyd.BaseModel):
    """Define a source for seed data for the newly created database"""

    type: _t.Literal["json", "sql", "file", "module"] = _pyd.Field(
        title="Type",
        description=_seeddata_type_desc,
    )
    value: str = _pyd.Field(
        title="Value",
        description=(
            "Value that corresponds to the type. See the type field for more information."
        ),
    )


class CreateDbArgs(_pyd.BaseModel):
    url: str = _pyd.Field(
        title="Connection string",
        description=(
            "A database connection string, eg: `postgres://user:password@localhost:5432/db_name`. "
            "The user/password combination must have database CREATE privileges."
        ),
    )
    append_name: str = _pyd.Field(
        default="",
        title="Append name",
        description=(
            "String to append to the database name when creating. Used for differentiating tests"
        ),
    )
    with_timestamp: bool = _pyd.Field(
        default=False,
        title="With timestamp",
        description=(
            'Set to "true" to include a timestamp in a created database name. This can help '
            "differentiate different runs of the same test"
        ),
    )
    schema_def: SchemaDef = _pyd.Field(
        alias="schema",
        title="Schema Definition",
        description=(
            "Used to create the database schema after the database has been created."
        ),
    )
    seeds: _t.List[SeedData] = _pyd.Field(
        default=[],
        title="Seed data",
        description=(
            "Seed data files to load into the database after creation. Will be loaded in the same "
            "order they are provided."
        ),
    )
    keep_db_on_error: bool = _pyd.Field(
        default=False,
        title="Keep Database on Error",
        description=(
            "By default an error while creating the database will cause the server to return the "
            "error raised and clean up the broken build. Setting this value to `true` will keep "
            "the broken database intact."
        ),
    )
    reset_seq: bool = _pyd.Field(
        default=False,
        title="Reset sequences",
        description=(
            "When using a SQLAlchemy module as the schema, any autoincrement sequences in the "
            "database will **not** be automatically set correctly to the maximum value of the "
            "inserted items. This can lead to insert problems later, where a newly inserted "
            "value will attempt to take an ID already taken by another item in the database. "
            "This option **only** works when using a sqlalchemy metadata object to load the "
            "schema, and will find all of the known autoincrementing sequences and set them equal "
            "to the highest value in the column that the sequence is attached to. It will do "
            "this only if seed values are loaded, and only if the schema was loaded via a "
            "sqlalchemy MetaData object. If loading via SQL just make sure to set the sequences "
            "properly as part of loading the seed."
        ),
    )


class DropDbArgs(_pyd.BaseModel):
    drop_id: str = _pyd.Field(
        title="Drop ID",
        description=(
            "The `drop_id` passed back to you from the initial database creation."
        ),
    )



import sqlalchemy as _sa
import sqlalchemy.engine as _sae
import sqlalchemy.engine.base as _saeb
from . import serialize as _dbu_ser, conn as _dbu_conn, schema as _dbu_schema
from .. import file as _fu
import logging as _logging
from ... import types as _types

_log = _logging.getLogger(__name__)


def _yield_all_tables(metadata: _sa.MetaData):
    yield from metadata.tables.values()


def _yield_all_columns(table: _sa.Table):
    yield from table.c


def reset_pg_seq(conn: _saeb.Connection, metadata: _sa.MetaData):
    """When loading data with postgresql any auto-incrementing counters are not automatically
    accounted for, so this will go through every column in every table, check if it has a
    counter, and if it does, set the value of that counter to the maximum value currently held
    within that column.
    """
    for table in _yield_all_tables(metadata):
        for col in _yield_all_columns(table):
            # check for sequence
            seq_id = (
                conn.execute(
                    f"SELECT pg_get_serial_sequence('{table.name}', '{col.name}')"
                )
                .scalars()
                .one_or_none()
            )
            if not seq_id:
                # no sequence attached to this column
                continue
            # find the highest current value in that table column
            max_val = conn.execute(_sa.select(_sa.func.max(col))).scalars().one()
            if max_val is None:
                # no values exist in this column; no need to change any sequences
                continue
            _log.info(f"Updating sequence {seq_id}; setting to {max_val}")
            new_max = (
                conn.execute(f"SELECT setval('{seq_id}', '{max_val}')").scalars().one()
            )
            if new_max != max_val:
                # something went wrong
                _log.error(
                    f"Failure setting {seq_id}; tried to set to {max_val}, received {new_max}"
                )


def load_json_seed(conn: _sae.Connection, seed: str):
    parsed = _fu.str_to_json(seed)
    metadata_str = parsed.get("metadata")
    if not metadata_str:
        raise _types.ApiError(
            message="JSON Seed data does not specify metadata"
        ).to_httperr()
    metadata = _dbu_schema.metadata_from_str(metadata_str)
    data = parsed.get("data")
    if not data:
        raise _types.ApiError(
            message="JSON Seed data does not specify any seed data"
        ).to_httperr()

    _dbu_ser.deserialize_db(
        metadata,
        parsed,
        conn=conn,
    )


def load_sql_seed(conn: _sae.Connection, seed: str):
    with conn.begin():
        conn.execute(seed)



from unittest import TestCase as _TestCase
import datetime as _dt
from . import url as _dbu

_TC = _TestCase()

_use_timestamp = _dt.datetime.strptime(
    "2022-04-23 14:05:34",
    "%Y-%m-%d %H:%M:%S",
)


def test_make_db_name():
    created_name = _dbu.make_db_name(
        "my_database",
        append="My appended string",
        at_timestamp=_use_timestamp,
    )
    _TC.assertEqual(
        created_name,
        "20220423140534_my_database_my_appended_string",
    )


def test_make_db_name_long():
    created_name = _dbu.make_db_name(
        "my_database",
        append="My appended and very long test name string",
        at_timestamp=_use_timestamp,
    )
    _TC.assertEqual(
        created_name,
        "20220423140534_my_database_my_appended_and_very_long_tes5119",
    )

def test_make_db_name_long_2():
    created_name = _dbu.make_db_name(
        "my_database",
        append="My appended and very long test name string thing",
        at_timestamp=_use_timestamp,
    )
    _TC.assertEqual(
        created_name,
        "20220423140534_my_database_my_appended_and_very_long_tesbf3f",
    )

def test_make_db_name_no_timestamp():
    created_name = _dbu.make_db_name(
        "my_database",
        append="My appended string",
    )
    _TC.assertEqual(
        created_name,
        "my_database_my_appended_string",
    )


def test_make_db_name_no_append():
    created_name = _dbu.make_db_name(
        "my_database",
        at_timestamp=_use_timestamp,
    )
    _TC.assertEqual(
        created_name,
        "20220423140534_my_database",
    )


def test_make_db_name_no_extra():
    created_name = _dbu.make_db_name(
        "my_database",
    )
    _TC.assertEqual(
        created_name,
        "my_database",
    )



import sqlalchemy as _sa
import sqlalchemy.engine as _sae
import typing as _t
from dateutil.parser import parse as _dateparse
import sqlalchemy.engine.base as _saeb
import sqlalchemy.orm as _sao
import sqlalchemy.types as _sat

import sqlalchemy as _sa

from .. import file as _fu

import logging as _logging

_log = _logging.getLogger(__name__)

SerializedTable = dict[str, _t.Any]


class SerializedDb(_t.TypedDict):
    """Type definition for a serialized data structure. The definition below leads to the basic
    structure:

    ```
    {
        "metadata": "path.to.module:metadata",
        "alembic": "<Alembic Revision>",
        "data": {
            "<table_name>": [
                {
                    "<column_name>": <value>,
                    ...
                },
                ...
            ],
            ...
        }
    }
    ```

    If `alembic` is not set, then the data cannot be tied to a specific
    revision of the database.
    """

    metadata: str
    # the alembic version number (if applicable)
    alembic: _t.Optional[str]
    # Actual table data. Takes the format of a dictionary of table names which map to lists of
    # objects which are dicts of column names to whatever SQLAlchemy data they contain.
    data: dict[str, list[SerializedTable]]


def _serialize_value(sql_type: _sat.TypeEngine, value: _t.Any) -> _t.Any:
    if hasattr(sql_type, "to_json"):
        return sql_type.to_json(value)
    elif isinstance(sql_type, _sa.DateTime):
        if value is not None:
            return value.isoformat()
        return value
    elif isinstance(sql_type, _sa.Enum):
        if value is not None:
            return value.name
        return value
    else:
        return value


def _deserialize_value(sql_type: _sat.TypeEngine, value: _t.Any) -> _t.Any:
    if hasattr(sql_type, "from_json"):
        return sql_type.from_json(value)
    elif isinstance(sql_type, _sa.DateTime):
        if value is not None:
            return _dateparse(value)
        return value
    else:
        return value


def serialize_db(
    metadata: _sa.MetaData, *, conn: _sao.Session, metadata_str: str
) -> SerializedDb:
    model_dict = dict()
    for t_name, t in metadata.tables.items():
        q = conn.query(t)
        all_rows = q.all()
        count_rows = q.count()
        _log.info(f"Dumping {count_rows} from table `{t_name}` ...")
        model_dict[t_name] = [
            {
                col.name: _serialize_value(col.type, val)
                for val, col in zip(r, t.columns.values())
            }
            for r in all_rows
        ]

    data_dict = {
        "metadata": metadata_str,
        "alembic": None,
        "data": model_dict,
    }

    return data_dict


def deserialize_db(
    metadata: _sa.MetaData,
    seed_dict: SerializedDb,
    *,
    conn: _sae.Connection,
) -> None:
    """Loads a model dictionary into the database."""
    model_dict = seed_dict["data"]
    for t_name in model_dict.keys():
        t = metadata.tables[t_name]
        _log.info(f"Loading {len(model_dict[t_name])} rows into table `{t_name}` ...")
        if len(model_dict[t_name]) <= 0:
            continue  # skip empty tables; sometimes trying to load them acts strangely
        rows = [
            {
                col_name: _deserialize_value(t.columns[col_name].type, val)
                for col_name, val in r.items()
            }
            for r in model_dict[t_name]
        ]
        conn.execute(t.insert(), [r for r in rows if len(r) > 0])



import unittest as _ut

_TC = _ut.TestCase()


def test_json_to_str():
    from .file import json_to_str

    mydict = {
        "key": 12,
        "15": "value",
    }
    json_str = json_to_str(mydict)
    expected = '{\n  "15": "value",\n  "key": 12\n}'
    _TC.assertEqual(json_str, expected)


def test_str_to_json():
    from .file import str_to_json

    json_str = '{ "key": 8092, "otherkey": "value", "59": 20 }'
    expected = {
        "key": 8092,
        "otherkey": "value",
        "59": 20,
    }
    json_dict = str_to_json(json_str)
    _TC.assertEqual(json_dict, expected)



import sqlalchemy as _sa
import sqlalchemy.engine as _sae
from .. import file as _fu
from ... import types as _types


def load_sqlalchemy_schema(url: _sae.URL | str, value: str):
    metadata = _fu.import_from_str(value)
    if not isinstance(metadata, _sa.MetaData):
        raise _types.ApiError(
            message=f"Not a valid sqlalchemy metadata object: {value}"
        ).to_httperr()
    engine = _sae.create_engine(url)
    metadata.create_all(engine)


def load_sql_schema(url: _sae.URL | str, value: str):
    engine = _sae.create_engine(url)
    with engine.connect() as conn:
        with conn.begin():
            conn.execute(value)


def metadata_from_str(metadata_str: str) -> _sa.MetaData:
    metadata = _fu.import_from_str(metadata_str)
    if not isinstance(metadata, _sa.MetaData):
        raise _types.ApiError(
            message=f"Not a valid metadata module: {metadata_str}"
        ).to_httperr()
    return metadata



import pytest
from .__test__.fixtures import (
    sldb_test_url,
    sldb_test_schema,
    sldb_test_schema_module,
    sldb_test_schema_def,
    sldb_test_seed_data,
    sldb_test_reset_seq,
    sldb_url,
    sldb_engine,
    sldb_conn,
)



import typing as _t
import unittest as _ut
import sqlalchemy_utils as _su
import sqlalchemy as _sa
import sqlalchemy.engine as _sae
import pytest as _pytest
from . import app as _app
from . import types as _types
from .__test__.db import models as _m
from .util import file as _file

_TC = _ut.TestCase()


def test_create_drop_test_db(sldb_test_url, sldb_test_schema_module):
    response = _app.create_db(
        _types.CreateDbArgs(
            url=sldb_test_url,
            append_name="test append",
            with_timestamp=True,
            schema=_types.SchemaDef(type="sqlalchemy", value=sldb_test_schema_module),
            seeds=[
                _types.SeedData(
                    type="module", value="sl.dbserver.__test__:seeds/test01.json"
                )
            ],
            reset_seq=True,
        )
    )
    _TC.assertTrue(_su.database_exists(response.url))

    # check that the tables were created
    engine = _sae.create_engine(response.url)
    with engine.connect() as connection:
        all_users = connection.execute(_sa.select(_m.User)).all()
    _TC.assertEqual(
        all_users,
        [
            (1, "user1", "user1@email.com"),
            (2, "user2", "user2@email.com"),
        ],
    )
    engine.dispose()

    # destroy the database, afterwards
    _app.drop_db(_types.DropDbArgs(drop_id=response.drop_id))
    _TC.assertFalse(_su.database_exists(response.url))


def test_reset_seq(sldb_conn: _sae.Connection):
    user_seq = sldb_conn.exec_driver_sql("SELECT last_value FROM user_id_seq").one()
    _TC.assertEqual(user_seq, (2,))


class TestNoResetSeq:
    @_pytest.fixture(scope="function")
    def sldb_test_reset_seq(self):
        return False

    def test_no_reset_seq(self, sldb_conn: _sae.Connection):
        user_seq = sldb_conn.exec_driver_sql("SELECT last_value FROM user_id_seq").one()
        _TC.assertEqual(user_seq, (1,))


class TestSqlLoad:
    @_pytest.fixture(scope="function")
    def sldb_test_seed_data(self):
        return [
            _types.SeedData(
                type="module", value="sl.dbserver.__test__:seeds/test01.sql"
            ),
        ]

    def test_load_sql_seed(self, sldb_conn):
        all_users = sldb_conn.execute(_sa.select(_m.User)).all()
        _TC.assertEqual(
            all_users,
            [
                (1, "user1", "user1@email.com"),
                (2, "user2", "user2@email.com"),
            ],
        )


class TestSqlSchema:
    @_pytest.fixture(scope="function")
    def sldb_test_schema_def(self) -> _types.SchemaDef:
        path_to_schema = _file.module_path("sl.dbserver.__test__:schema.sql")
        return _types.SchemaDef(type="file", value=path_to_schema)

    def test_load_sql_schema(self, sldb_conn):
        all_users = sldb_conn.execute(_sa.select(_m.User)).all()
        _TC.assertEqual(
            all_users,
            [
                (1, "user1", "user1@email.com"),
                (2, "user2", "user2@email.com"),
            ],
        )


class TestRawSqlSchema:
    @_pytest.fixture(scope="function")
    def sldb_test_schema_def(self) -> _types.SchemaDef:
        path_to_schema = _file.module_path("sl.dbserver.__test__:schema.sql")
        with open(path_to_schema, "r") as f:
            raw_sql = f.read()
        return _types.SchemaDef(type="raw_sql", value=raw_sql)

    def test_load_sql_schema(self, sldb_conn):
        all_users = sldb_conn.execute(_sa.select(_m.User)).all()
        _TC.assertEqual(
            all_users,
            [
                (1, "user1", "user1@email.com"),
                (2, "user2", "user2@email.com"),
            ],
        )



import sqlalchemy.engine as _sae
import sqlalchemy.exc as _saex
import hashlib as _hashlib
import datetime as _dt
import slugify as _slug
from ... import types as _types


def truncate_for_postgres(name: str) -> str:
    """Postgres can only have a maximum identifier length of 63 characters, so this will take
    a passed-in name and truncate it with an MD5 hash based on the full name appended to the end.
    This is the same way that postgres will rename a too-long identifier when it is created
    under the hood. Postgres seems to cap these shortened names at 60 characters, though, rather
    than taking up the full allowed 63.
    """
    # postgres uses first five characters of md5 of the full string
    if len(name) < 63:
        return name
    hashed = _hashlib.md5(name.encode()).hexdigest()[-4:]
    return name[: 60 - len(hashed)] + hashed


def make_db_name(
    base_name: str, *, append: str = "", at_timestamp: _dt.datetime | None = None
) -> str:
    """Constructs a new database name from a base name, appending a string and a timestamp,
    if provided.

    It will take the format:

    <timestamp>-<base_name>-<append>

    If `timestamp` or `append` are omitted, then that field and the adjacent dash will also
    be omitted. The "append" name is also run through a slugify so may not be verbatim the same
    as it was submitted.

    The default build of Postgresql has trouble with names over 63 characters, and will
    truncate them with a hash at the end. This will also do the same when creating the
    database name, regardless of what database engine is being used.

    Timestamp comes first to ensure that, if the base database name is long enough to trigger
    the postgres shortening algorithm, the timestamp will still be present so the user can
    know which databases were created most recently.
    """
    new_name = "_".join(
        [
            *(
                [at_timestamp.strftime("%Y%m%d%H%M%S")]
                if at_timestamp is not None
                else []
            ),
            base_name,
            *([_slug.slugify(append, separator="_")] if append else []),
        ]
    )
    return truncate_for_postgres(new_name)


def make_url(conn_str: str) -> _sae.URL | _types.ApiError:
    """Light wrapper around SQLAlchemy's URL parser to catch and return a standardized error
    type.
    """
    try:
        return _sae.make_url(conn_str)
    except _saex.ArgumentError:
        return _types.ApiError(message="Invalid URL")



import typing as _t
import sqlalchemy_utils as _su
import sqlalchemy as _sa
import sqlalchemy.orm as _sao
import sqlalchemy.engine as _sae
import sqlalchemy.engine.base as _saeb
import contextlib as _cl
import logging as _logging

_log = _logging.getLogger(__name__)


def _disable_constraints(dialect: str, sqlFn: _t.Callable[[str], None]):
    """Given the dialect type and a function that can take SQL statements and execute them,
    will disable the check constraints for the given dialect type.
    """
    if dialect == "sqlite":
        sqlFn("PRAGMA ignore_check_constraints=OFF")
        sqlFn("PRAGMA foreign_keys=OFF")
    elif dialect == "postgresql":
        # Only works when constraints are configured to be deferrable
        # sqlFn('SET CONSTRAINTS ALL DEFERRED')
        # More dangerous (because constraints are straight up not checked) but ideal for
        # restoring files from backup
        sqlFn("SET session_replication_role = replica")


def _enable_constraints(dialect: str, sqlFn: _t.Callable[[str], None]):
    """Given the dialect type and a function that will run sql statements, re-enables the check
    constraints for the given dialect type
    """
    if dialect == "sqlite":
        sqlFn("PRAGMA ignore_check_constraints=ON")
        sqlFn("PRAGMA foreign_keys=ON")
    elif dialect == "postgresql":
        # Only works when constraints are configured to be deferrable
        # sqlFn('SET CONSTRAINTS ALL IMMEDIATE')
        sqlFn("SET session_replication_role = DEFAULT")


@_cl.contextmanager
def disabled_constraints(engine: _saeb.Engine):
    dialect = engine.dialect.name

    with engine.connect() as conn:

        def run_sql(arg: str):
            nonlocal conn
            conn.exec_driver_sql(arg)

        _disable_constraints(dialect, run_sql)
        yield conn
        _enable_constraints(dialect, run_sql)


def disable_session_constraints(session: _sao.Session):
    dialect = session.get_bind().dialect.name

    def run_sql(arg: str):
        nonlocal session
        session.execute(arg)

    _disable_constraints(dialect, run_sql)


def enable_session_constraints(session: _sao.Session):
    dialect = session.get_bind().dialect.name

    def run_sql(arg: str):
        nonlocal session
        session.execute(arg)

    _enable_constraints(dialect, run_sql)


def drop_database(conn_str: _sae.URL | str):
    """Uses the provided connection string to ensure that the database it points to does not
    exist.
    """
    if _su.database_exists(str(conn_str)):
        _su.drop_database(str(conn_str))


def create_database(conn_str: _sae.URL | str):
    """Uses the provided connection string to ensure that the database it points to exists.
    If the database already exists, then it will destroy and recreate it.
    """
    drop_database(conn_str)
    if not _su.database_exists(str(conn_str)):
        _su.create_database(str(conn_str))



import typing as _t
import sqlalchemy as _sa
import sqlalchemy.engine as _sae
import pathlib as _pl
from .util import file as _fu
from . import types as _types
from .util.db import (
    schema as _dbu_schema,
    seed as _dbu_seed,
    conn as _dbu_conn,
)


def create_schema(url: _sae.URL | str, schema: _types.SchemaDef):
    """Uses the schema definition to create the schema inside the passed database url"""
    match schema.type:
        case "sqlalchemy":
            _dbu_schema.load_sqlalchemy_schema(url, schema.value)
        case "file":
            with open(schema.value, "r") as f:
                fdata = f.read()
            _dbu_schema.load_sql_schema(url, fdata)
        case "raw_sql":
            _dbu_schema.load_sql_schema(url, schema.value)


def _seed_data_from_file(fname: str) -> _types.SeedData:
    fext = _pl.Path(fname).suffix.lower()
    match fext:
        case ".json":
            ftype = "json"
        case ".sql":
            ftype = "sql"
        case _:
            raise _types.ApiError(
                message=f"File type extension not recognized, must be `sql` or `json`: {fname}"
            ).to_httperr()
    with open(fname, "r") as f:
        fdata = f.read()

    return _types.SeedData(type=ftype, value=fdata)


def _seed_data_from_module(modname: str) -> _types.SeedData:
    try:
        fname = _fu.module_path(modname)
    except ValueError as e:
        raise _types.ApiError(message=str(e)).to_httperr()
    return _seed_data_from_file(fname)


def load_seed_data_list(
    url: _sae.URL | str,
    seeds: _t.List[_types.SeedData],
    *,
    reset_seq: bool,
    schema: _types.SchemaDef,
):
    if len(seeds) <= 0:
        return
    engine = _sae.create_engine(url)
    with _dbu_conn.disabled_constraints(engine) as conn:
        for seed in seeds:
            load_seed_data(conn, seed)
        if reset_seq:
            _reset_seq(conn, schema)


def load_seed_data(conn: _sae.Connection, seed: _types.SeedData):
    if seed.type == "file":
        seed = _seed_data_from_file(seed.value)
    if seed.type == "module":
        seed = _seed_data_from_module(seed.value)

    # at this point seed should have either a "json" or "sql" type
    match seed.type:
        case "json":
            _dbu_seed.load_json_seed(conn, seed.value)
        case "sql":
            _dbu_seed.load_sql_seed(conn, seed.value)


def _reset_seq(conn: _sae.Connection, schema: _types.SchemaDef):
    if schema.type != "sqlalchemy":
        # nothing to do
        return

    metadata = _fu.import_from_str(schema.value)
    if not metadata or not isinstance(metadata, _sa.MetaData):
        raise _types.ApiError(
            message=f"Not a valid metadata module: {schema.value}"
        ).to_httperr()

    _dbu_seed.reset_pg_seq(conn, metadata)



import pytest as _pytest
import typing as _t
import sqlalchemy as _sa
import sqlalchemy.engine as _sae
from .db import models as _m
from .. import (
    app as _app,
    types as _types,
)


@_pytest.fixture(scope="function")
def sldb_test_schema() -> _sa.MetaData:
    return _m.metadata


@_pytest.fixture(scope="function")
def sldb_test_url(pytestconfig) -> str:
    """The test connection to use for this test run"""
    db_url = pytestconfig.getoption("--testdb")
    if db_url is None:
        db_url = pytestconfig.getini("testdb")
    return db_url


@_pytest.fixture(scope="function")
def sldb_test_schema_module() -> str:
    return "sl.dbserver.__test__.db.models:metadata"


@_pytest.fixture(scope="function")
def sldb_test_schema_def(sldb_test_schema_module) -> _types.SchemaDef:
    return _types.SchemaDef(type="sqlalchemy", value=sldb_test_schema_module)


@_pytest.fixture(scope="function")
def sldb_test_seed_data() -> _t.List[_types.SeedData]:
    return [
        _types.SeedData(type="module", value="sl.dbserver.__test__:seeds/test01.json")
    ]


@_pytest.fixture(scope="function")
def sldb_test_reset_seq() -> bool:
    return True


@_pytest.fixture(scope="function")
def sldb_url(
    sldb_test_url,
    sldb_test_schema_def,
    request,
    sldb_test_seed_data,
    sldb_test_reset_seq,
):
    """Creates, yields, and cleans up a testing database"""
    test_name = request.node.name
    response = _app.create_db(
        _types.CreateDbArgs(
            url=sldb_test_url,
            append_name=test_name,
            with_timestamp=True,
            schema=sldb_test_schema_def,
            seeds=sldb_test_seed_data,
            reset_seq=sldb_test_reset_seq,
        )
    )
    yield response.url
    _app.drop_db(_types.DropDbArgs(drop_id=response.drop_id))


@_pytest.fixture(scope="function")
def sldb_engine(sldb_url):
    engine = _sae.create_engine(sldb_url)
    yield engine
    engine.dispose()


@_pytest.fixture(scope="function")
def sldb_conn(sldb_engine):
    with sldb_engine.connect() as conn:
        yield conn



from unittest import TestCase as _TestCase
from . import db as _db

_TC = _TestCase()


def test_seed_data_from_module_json():
    loaded = _db._seed_data_from_module("sl.dbserver.__test__:seeds/test01.json")
    _TC.assertEqual(loaded.type, "json")


def test_seed_data_from_module_sql():
    loaded = _db._seed_data_from_module("sl.dbserver.__test__:seeds/test01.sql")



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.md")) as f:
    README = f.read()

requires = [
    # For DB management
    "sqlalchemy[asyncio]          >= 1, < 2",
    "python-dateutil              >= 2",
    "sqlalchemy-utils",
    "psycopg2-binary              >= 2",
    "python-slugify               >= 4",
    # For server
    "fastapi[all]",
    "uvicorn[standard]",
]

dev_requires = [
    "pytest                >= 7",
    "pytest-cov            >= 4",
    "pytest-asyncio        >= 0.17",
    "mypy                  >= 0.812",
    "requests              >= 2, < 3",
    "types-python-dateutil >= 2, < 3",
    "black                 >= 23",
    "alembic               >= 1",
    "build",
    "twine",
]

setup(
    name="sl-dbserver",
    version="0.0.1a3",
    license="MIT",
    description="Springless Database Server",
    long_description=README,
    author="cogs",
    author_email="cogs@springless.com",
    url="https://github.com/Springless/sl-dbserver",
    packages=find_packages("src"),
    package_dir={"": "src"},
    namespace_packages=["sl"],
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    extras_require={
        "dev": dev_requires,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Testing",
    ],
    entry_points={
        "console_scripts": [
            "sl-dbserver=sl.dbserver.cli:start_server",
        ],
    },
    keywords="sl springless database server testing postgresql",
)



import fastapi as _fapi
import datetime as _dt
from .util.db import (
    url as _dbu_url,
    conn as _dbu_conn,
)
from . import (
    types as _types,
    db as _db,
)


def create_db(args: _types.CreateDbArgs = _fapi.Body()) -> _types.CreatedDb:
    """Create a new database based on the passed-in connection string and options."""
    passed_url = _dbu_url.make_url(args.url)
    if isinstance(passed_url, _types.ApiError):
        raise passed_url.to_httperr()

    if not passed_url.database:
        raise Exception("No database found in parsed URL string")

    new_url = passed_url.set(
        database=_dbu_url.make_db_name(
            passed_url.database,
            append=args.append_name or "",
            at_timestamp=_dt.datetime.now() if args.with_timestamp else None,
        )
    )

    try:
        _dbu_conn.create_database(new_url)
        _db.create_schema(new_url, args.schema_def)
        _db.load_seed_data_list(
            new_url,
            args.seeds,
            reset_seq=args.reset_seq,
            schema=args.schema_def,
        )
    except Exception as e:
        if not args.keep_db_on_error:
            _dbu_conn.drop_database(new_url)
        raise e

    return _types.CreatedDb(url=str(new_url), drop_id=str(new_url))


def drop_db(args: _types.DropDbArgs = _fapi.Body()) -> None:
    """Drop a created database"""
    passed_url = _dbu_url.make_url(args.drop_id)

    if isinstance(passed_url, _types.ApiError):
        raise passed_url.to_httperr()

    if not passed_url.database:
        raise _types.ApiError(
            message="No database found in parsed URL string"
        ).to_httperr()

    _dbu_conn.drop_database(passed_url)

    return None


def build_app():
    app = _fapi.FastAPI()

    app.post("/create")(create_db)
    app.post("/drop")(drop_db)

    return app


main = build_app()



import sqlalchemy as _sa
import sqlalchemy.sql.schema as _sass

# Recommended naming convention used by Alembic, as various different database
# providers will autogenerate vastly different names making migrations more
# difficult. See: http://alembic.zzzcomputing.com/en/latest/naming.html
NAMING_CONVENTION = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}

metadata = _sass.MetaData(naming_convention=NAMING_CONVENTION)

User = _sa.Table(
    "user",
    metadata,
    _sa.Column("id", _sa.Integer, primary_key=True),
    _sa.Column("name", _sa.String, nullable=False),
    _sa.Column("email", _sa.String, nullable=True, unique=True),
)

Post = _sa.Table(
    "post",
    metadata,
    _sa.Column("id", _sa.Integer, primary_key=True),
    _sa.Column("user_id", _sa.Integer, _sa.ForeignKey("user.id"), nullable=False),
    _sa.Column(
        "timestamp",
        _sa.DateTime(timezone=True),
        nullable=False,
        server_default=_sa.func.now(),
    ),
    _sa.Column("body", _sa.String, nullable=False),
)
